from typing import List, Union, Optional

from transformers import T5Tokenizer
import re

def _upper_repl(matchobj):
    #print(matchobj, matchobj.group(0))
    return '<cZ>' + matchobj.group(0).lower()

def _cap_repl(matchobj):
    #print(matchobj, matchobj.group(0))
    return matchobj.group(0)[4:].upper()

_UpperPattern = re.compile('([A-Z][a-z])')
_CapitalizedPattern = re.compile(r'(\<cZ\>[a-z])')

def pre_encode(s):
    if isinstance(s, str):
        s = _UpperPattern.sub(_upper_repl, s)
        return s.replace('\t', '    ').replace('\n', '<nL>')
    if isinstance(s, tuple):
        return tuple(map(pre_encode, s))
    return s

def post_decode(s):
    if isinstance(s, str):
        return _CapitalizedPattern.sub(_cap_repl,s).replace('<nL>', '\n')
    return s


class PapertownTokenizer(T5Tokenizer):

    def tokenize(self, text: str, pair: Optional[str] = None, add_special_tokens: bool = False) -> List[str]:
        text=pre_encode(text)
        pair=pre_encode(pair)
        return super.tokenizer(text, add_special_tokens=add_special_tokens)

    def encode_plus(self, 
        text,
        text_pair = None,
        add_special_tokens: bool = True,
        padding = False,
        truncation = None,
        max_length: Optional[int] = None,
        stride: int = 0,
        is_split_into_words: bool = False,
        pad_to_multiple_of: Optional[int] = None,
        return_tensors = None,
        return_token_type_ids: Optional[bool] = None,
        return_attention_mask: Optional[bool] = None,
        return_overflowing_tokens: bool = False,
        return_special_tokens_mask: bool = False,
        return_offsets_mapping: bool = False,
        return_length: bool = False,
        verbose: bool = True,
        **kwargs,
    ):
        text = pre_encode(text)
        text_pair=pre_encode(text_pair)
        return super().encode_plus(text, 
            text_pair=text_pair,
            add_special_tokens = add_special_tokens,
            padding=padding,
            truncation=truncation,
            max_length=max_length,
            stride=stride,
            is_split_into_words=is_split_into_words,
            pad_to_multiple_of=pad_to_multiple_of,
            return_tensors=return_tensors,
            return_token_type_ids=return_token_type_ids,
            return_attention_mask=return_attention_mask,
            return_overflowing_tokens=return_overflowing_tokens,
            return_special_tokens_mask=return_special_tokens_mask,
            return_offsets_mapping=return_offsets_mapping,
            return_length=return_length,
            verbose=verbose,
            **kwargs,
    )

    def batch_encode_plus(self,
        batch_text_or_text_pairs,
        add_special_tokens: bool = True,
        padding = False,
        truncation = None,
        max_length = None,
        stride: int = 0,
        is_split_into_words: bool = False,
        pad_to_multiple_of: Optional[int] = None,
        return_tensors = None,
        return_token_type_ids: Optional[bool] = None,
        return_attention_mask: Optional[bool] = None,
        return_overflowing_tokens: bool = False,
        return_special_tokens_mask: bool = False,
        return_offsets_mapping: bool = False,
        return_length: bool = False,
        verbose: bool = True,
        **kwargs,
    ):
        batch_text_or_text_pairs = [pre_encode(x) for x in batch_text_or_text_pairs]        
        return super().batch_encode_plus(
            batch_text_or_text_pairs=batch_text_or_text_pairs,
            add_special_tokens=add_special_tokens,
            padding=padding,
            truncation=truncation,
            max_length=max_length,
            stride=stride,
            is_split_into_words=is_split_into_words,
            pad_to_multiple_of=pad_to_multiple_of,
            return_tensors=return_tensors,
            return_token_type_ids=return_token_type_ids,
            return_attention_mask=return_attention_mask,
            return_overflowing_tokens=return_overflowing_tokens,
            return_special_tokens_mask=return_special_tokens_mask,
            return_offsets_mapping=return_offsets_mapping,
            return_length=return_length,
            verbose=verbose,
            **kwargs,
        )

    def onvert_tokens_to_string(self, tokens: List[str]):
        s = super().convert_tokens_to_string(tokens)
        return post_decode(s)
    
    def decode(self,
                         token_ids, 
                         skip_special_tokens: bool = False,
                         clean_up_tokenization_spaces: bool = None,
                        **kwargs):
        s = super().decode(token_ids, 
                        skip_special_tokens=skip_special_tokens,
                        clean_up_tokenization_spaces=clean_up_tokenization_spaces,
                        **kwargs)
        return post_decode(s)



import torch

def get_rank():
    if torch.distributed.is_available() and torch.distributed.is_initialized():
        rank = torch.distributed.get_rank()
    else:
        rank = 0


def train_model(DataComposer, model, args: TrainingArguments, use_flash=False):

    if use_flash:
        try:
            from optimum.bettertransformer import BetterTransformer
            model = BetterTransformer.transform(model)
        except:
            use_flash=False

    args = TrainingArguments(
        output_dir="./output",
        overwrite_output_dir=True,
        per_device_train_batch_size=128,
        per_device_eval_batch_size=128,
        auto_find_batch_size=True,  # バッチサイズ自動
        do_eval=False,
        logging_steps=1000,
#        gradient_accumulation_steps=1,
        num_train_epochs=1,
        weight_decay=0.1,
#        warmup_steps=1_000,
        lr_scheduler_type="cosine",
        learning_rate=5e-4, #TODO: 論文から探す
#        save_steps=5_000,
        fp16=use_fp16,
    )

    trainer = Trainer(
        model=model,
        args=args,
        train_dataset=dummy_dataset(max_length),
        data_collator=data_collator,
    )
    result = trainer.train()
    print_summary(result, use_flash)
    # モデルを保存 output_path に保存します
    if use_flash:
        model = BetterTransformer.reverse(model)
    output_path='trained'
    tokenizer.save_pretrained(output_path)
    model.save_pretrained(output_path)
    model.cpu()
    gc.collect()
    torch.cuda.empty_cache()
    print_gpu_utilization()



from typing import List
import random
import json
import numpy as np
import pandas as pd

import hashlib
from transformers import AutoTokenizer

from .tokenizers import load_tokenizer, find_ellipsis_token_id, find_newline_token_id, find_token_id
from .commons import *
from .file_utils import *

DEFAULT_BLOCK_SIZE=2048

def _record_tokens(counts):
    if len(counts) == 0:
        return {'total': 0}
    data = np.array(counts)
    return {
        'total': int(np.sum(data)),
        'mean': float(np.mean(data)),
        'std': float(np.var(data)) ** 0.5,
        'max': int(np.max(data)),
        '75%': int(np.percentile(data, 75)),
        'median': int(np.median(data)),
        '25%': int(np.percentile(data, 25)),
        'min': int(np.min(data)),
    }

def _update_fn(blocks: List[List[int]]):
    return blocks


empty_tokens = []


class DefaultSplitter(object):

    def __init__(self, tokenizer, block_size, **kwargs):
        self.tokenizer = tokenizer
        self.block_size = block_size
        self.split_args = kwargs
        self.pad_token_id = getint(kwargs, 'pad_token_id', tokenizer.pad_token_id)
        self.eos_token_id = getint(kwargs, 'eos_token_id', tokenizer.eos_token_id)
        self.ellipsis_token_id = find_ellipsis_token_id(tokenizer)
        # self.output_token_id = None # <sftの場合>
        self.trancate_size=getint(kwargs, 'trancate_size', 0)
        self.sep = kwargs.get('sep', None)
        self.token_counts = []
        self.text_length_count = 0
        self.trimmed_size = 0
        self.padded_size = 0
    
    def about(self):
        return dict(name=self.__class__.__name__)

    def split(self, text:str, blocks: List[List[int]]):
        raise NotImplemented()

    def flush(self, blocks: List[List[int]]):
        pass

    def split_iter(self, iterator, update_fn=_update_fn):
        blocks = []
        for text in iterator:
            self.split(text, blocks)
            blocks = update_fn(blocks)
        self.flush(blocks)
        blocks = update_fn(blocks)
        return blocks

    def encode_and_count(self, text):
        self.text_length_count += len(text)
        tokens = self.tokenizer.encode(text)
        self.token_counts.append(len(tokens)-1)
        return tokens

    def resize_token_counts(self, size):
        self.token_counts[-1] += size

    def report(self, logs: dict = None, verbose=True):
        token_count = sum(self.token_counts)
        if logs:
            logs['n_tokens'] = token_count
            if self.text_length_count > 0:
                logs['n_chars'] = self.text_length_count
                logs['tokens_per_char'] = token_count / self.text_length_count
            logs['tokens'] = _record_tokens(self.token_counts)
        if verbose:
            print(pd.DataFrame({'tokens': self.token_counts}).describe())
        if self.padded_size > 0:
            if verbose:
                print(f'padding: {self.padded_size:,} {self.padded_size*100/token_count:.2f}%')
            if logs:
                logs['padding_rate'] = self.padded_size / token_count


class SimpleTextBlockSplitter(DefaultSplitter):
    def __init__(self, tokenizer, block_size, **kwargs):
        super().__init__(tokenizer, block_size, **kwargs)
        self.extra_tokens=empty_tokens
        self.prefix='pre'

    def split(self, text:str, blocks: List[List[int]]):
        tokens = self.encode_and_count(text)
        work_size = self.block_size
        tokens = self.extra_tokens + tokens
        for i in range(0, len(tokens) - work_size + 1, work_size):  
            segmented = tokens[i : i + work_size]
            blocks.append(segmented)
        extra_size = len(tokens) % work_size
        if extra_size == 0:
            self.extra_tokens = empty_tokens
        else:
            self.extra_tokens = tokens[-extra_size:]

    def report(self, logs: dict = None, verbose=True):
        super().report(logs, verbose=verbose)
        if logs:
            logs['block_size'] = self.block_size


class MultiTextBlockSplitter(DefaultSplitter):
    def __init__(self, tokenizer, block_size, **kwargs):
        super().__init__(tokenizer, block_size, **kwargs)
        self.prefix='pre'
        self.work_size = getint(kwargs, 'work_size', 512)
        self.pad_size = getint(kwargs, 'pad_size', 64)
        # 警告が出る場合があるので、padding を変更する
        self.pad_token_id = find_newline_token_id(tokenizer)
        self.extra_tokens = empty_tokens
        self.work_buffers = []
        self.lazy_buffers = []
        self.heads = [[] for _ in range(self.work_size//self.pad_size)]
        self.match_count = 0
        self.sep_counts = []

    def resize_as_dividable(self, tokens, size, trancate_size=0, padding_size=None):
        extra_size = len(tokens) % size 
        if extra_size == 0:
            return tokens
        if extra_size <= trancate_size:
            tokens = tokens[:-extra_size]
            if len(tokens)>2:
                tokens[-1] = self.eos_token_id
                if self.ellipsis_token_id:
                    tokens[-2] = self.ellipsis_token_id
            return tokens
        if padding_size is None or (size - extra_size) < padding_size:
            self.padded_size += (size - extra_size)
            return tokens + [self.pad_token_id] * (size - extra_size)
        return tokens[:(len(tokens)//size)*len(tokens)]

    def tokenize_nonsep(self, text:str):
        tokens = self.tokenizer.encode(text)
        self.token_counts.append(len(tokens))
        tokens = self.resize_as_dividable(tokens, self.pad_size, trancate_size=self.trancate_size)
        return tokens

    def tokenize_sep(self, text:str):
        text_blocks = text.split(self.sep)
        tokenizer = self.tokenizer
        chunks = [tokenizer.convert_tokens_to_ids(tokenizer.tokenize(line)) for line in text_blocks]
        chunks[-1] = tokenizer.build_inputs_with_special_tokens(chunks[-1])
        counts = [len(c) for c in chunks]
        self.token_counts.append(sum(counts))
        self.sep_counts.extend(counts)
        tokens = []
        chunk_bufs = []
        split_size = self.pad_size * 2
        for chunk in chunks:
            prev_len = len(chunk_bufs)
            chunk_bufs.extend(chunk)
            if len(chunk_bufs) < split_size:
                continue
            if len(chunk_bufs) % self.pad_size == 0:
                tokens.extend(chunk_bufs)
                chunk_bufs=[]
            else:
                splitted = self.resize_as_dividable(chunk_bufs, self.pad_size, trancate_size=self.pad_size, padding_size=0)
                tokens.extend(splitted)
                if prev_len > 0:
                    chunk_bufs = chunk
                else:
                    chunk_bufs = []
        return self.resize_as_dividable(tokens+chunk_bufs, self.pad_size, self.trancate_size)

    def add_buffer(self, blocks: List[List[int]], tokens: List[int]):
        assert(len(tokens) == self.work_size)
        if len(self.extra_tokens) > 0:
            self.lazy_buffers.append(tokens)
        elif len(self.lazy_buffers) > 0:
            lazy_buffers = self.lazy_buffers
            self.lazy_buffers = []
            for lazy_tokens in lazy_buffers:
                self.add_buffer(blocks, lazy_tokens)

        self.work_buffers.extend(tokens)
        if len(self.work_buffers) == self.block_size:
            blocks.append(self.work_buffers)
            self.work_buffers = []

    def push_head(self, tokens):
        index = len(tokens) // self.pad_size
        self.heads[index].append(tokens)

    def pop_head(self, extra_size):
        index = extra_size // self.pad_size
        if len(self.heads[index]) > 0:
            return self.heads[index].pop()
        return None

    def split(self, text:str, blocks: List[List[int]]):
        if self.sep is not None and self.sep in text:
            tokens = self.tokenize_sep(text)
        else:
            tokens = self.tokenize_nonsep(text)

        work_size = self.work_size
        if len(tokens) < work_size:
            head = self.pop_head(work_size - len(tokens) % work_size)
            if head:
                self.add_buffer(blocks, head+tokens)
            else:
                self.push_head(tokens)
            return
        tokens = self.extra_tokens + tokens
        for i in range(0, len(tokens) - work_size + 1, work_size):  
            segmented = tokens[i : i + work_size]
            self.add_buffer(blocks, segmented)

        extra_size = len(tokens) % work_size
        if extra_size == 0: # 最後の分割が揃っていればおしまい
            self.extra_tokens = empty_tokens
            return
        extra_tokens = tokens[-extra_size:]
        head = self.pop_head(work_size - extra_size)
        if head:
            self.add_buffer(blocks, extra_tokens+head)
            self.extra_tokens = empty_tokens
        else:
            self.extra_tokens = extra_tokens

    def flush(self, blocks: List[List[int]]):
        heads = []
        for hh in self.heads:
            heads.extend(hh)
        # print(pd.DataFrame({'heads': [len(h) for h in heads]}).describe())
        random.shuffle(heads)
        tokens = []
        for t in [self.extra_tokens]+heads:
            tokens.append(t)
        work_size = self.work_size
        for i in range(0, len(tokens) - work_size + 1, work_size):  
            segmented = tokens[i : i + work_size]
            self.add_buffer(blocks, segmented)

    def report(self, logs: dict = None, verbose=True):
        super().report(logs, verbose=verbose)
        if logs:
            logs['block_size'] = self.block_size
            logs['work_size'] = self.work_size

class TextPairSplitter(DefaultSplitter):
    def __init__(self, tokenizer, block_size, **kwargs):
        super().__init__(tokenizer, block_size, **kwargs)
        self.prefix=''
        self.output_sep_token_id = find_token_id(tokenizer, 
                                                 kwargs.get('output_sep', '<outpuT>'), 
                                                 kwargs.get('sep', '<seP>'), 
                                                 '<nL>')

    def report(self, logs: dict = None, verbose=True):
        super().report(logs, verbose=verbose)
        if self.output_sep_token_id is not None:
            logs['output_sep_token_id'] = self.output_sep_token_id            
        if self.sep is not None:
            logs['corpus_sep'] = self.sep            

    def trancate_pair(self, inputs: List[int], labels: List[int], blocks: List[List[int]]):
        if self.block_size is not None:
            if len(labels) > self.block_size:
                # ラベルの方が大きい場合は諦める
                return
            if len(inputs)+len(labels) > self.block_size:
                trimmed_size = self.block_size - len(labels)
                if len(inputs) - trimmed_size > self.trancate_size:
                    # 諦める                
                    return
                self.trimmed_size += len(inputs) - trimmed_size
                inputs = inputs[:trimmed_size]
            else:
                self.padded_size += self.block_size - (len(inputs)+len(labels))
        index = len(inputs)
        inputs[-1] = self.output_sep_token_id
        blocks.append(inputs+labels+[index])

    def split(self, text:str, blocks: List[List[int]]):
        t = text.split(self.sep)
        if len(t)==2:
            inputs = self.encode_and_count(t[0])        
            labels = self.encode_and_count(t[1])
            self.trancate_pair(inputs, labels, blocks)
        else:
            print(self.sep, text)
            raise ValueError(f'In text, the {self.sep} token is required.')

class SimpleTextSplitter(TextPairSplitter):
    def __init__(self, tokenizer, block_size, **kwargs):
        super().__init__(tokenizer, block_size, **kwargs)
        # セパレータ字句があるか調べる
        # self.sep: コーパス上のセパレータ
        # self.output_sep トークンナイザのセパレータ
        self.output_sep_token_id = None
        self.output_sep = str(kwargs.get('output_sep', self.sep))
        if self.output_sep:
            self.output_sep_token_id = find_token_id(tokenizer, self.output_sep)
            if self.output_sep_token_id == tokenizer.unk_token_id:
                verbose_print(f'undefined token {self.output_sep} in {tokenizer.name_or_path}')
                self.output_sep_token_id = None            

    def trancate_text(self, inputs: List[int], labels: List[int], blocks: List[List[int]]):
        if self.block_size is not None:
            inputs_size = len(inputs)
            if inputs_size > self.block_size:
                if inputs_size - self.block_size > self.trancate_size:
                    # 切り詰めが大きすぎる
                    return
                half_size = self.block_size // 2
                prefix = inputs[:half_size]
                suffix = inputs[-half_size:]
                if self.ellipsis_token_id:
                    prefix[-1] = self.ellipsis_token_id
                inputs = prefix + suffix
                self.trimmed_size += (inputs_size - len(inputs))
            else:
                self.padded_size += self.block_size - inputs_size
        blocks.append(inputs)

    def split(self, text:str, blocks: List[List[int]]):
        if self.sep:
            text = text.replace(self.sep, self.output_sep, 1)
        inputs = self.encode_and_count(text)
        if self.output_sep_token_id is not None:
            index = inputs.find(self.output_sep_token_id)
            if index == -1:
                sep = self.sep or self.output_sep
                raise ValueError(f'{sep} が見つかりません')
            self.trancate_pair(inputs[:index+1], inputs[index+1:], blocks)
        else:
            self.trancate_text(inputs, blocks)

def new_TextSplitter(tokenizer, training_type, format='simple', block_size=None, **kwargs):
    splitter = None
    if training_type.startswith('pre'):
        if block_size is None:
            verbose_print(f"block_sizeの指定がないため、block_size={DEFAULT_BLOCK_SIZE}にします。")
            block_size = DEFAULT_BLOCK_SIZE
        if format=='multi':
            splitter = MultiTextBlockSplitter(tokenizer, block_size, **kwargs)
        if splitter is None:
            if format != 'simple':
                verbose_print(f"format={format}は、サポートされていません。")
        splitter = SimpleTextBlockSplitter(tokenizer, block_size, **kwargs)
    elif training_type.startswith('seq2seq'):
        # if splitter is None:
        #     if format != 'simple':
        #         verbose_print(f"format={format}は、サポートされていません。")
        if 'sep' not in kwargs:
            kwargs['sep'] = '<outpuT>'
        splitter = TextPairSplitter(tokenizer, block_size, **kwargs)
    else: # ファインチューニング用
        # if splitter is None:
        #     if format != 'simple':
        #         verbose_print(f"format={format}は、サポートされていません。")
        splitter = SimpleTextSplitter(tokenizer, block_size, **kwargs)
    return splitter


## Store 用

def record_tokenizer(tokenizer: AutoTokenizer):
    allvoc = ''.join(tokenizer.get_vocab().keys())
    sha256 = hashlib.sha256(allvoc.encode()).hexdigest()
    return dict(
        name_or_path=tokenizer.name_or_path,
        pad_token_id = tokenizer.pad_token_id,
        eos_token_id = tokenizer.eos_token_id,
        hash=sha256, 
        vocab_size=tokenizer.vocab_size)


N_CHUNKS = 4096

class DatasetStore(object):
    def __init__(self, store_path, prefix, block_size, **kwargs):
        self.store_path = safe_dir(store_path)
        self.prefix = prefix
        self.block_size = block_size
        self.config = {}
        self.file_ext = kwargs.get("file_ext", "npz")
        self.n_chunks = kwargs.get("n_chunks", N_CHUNKS)
        self.shuffle = kwargs.get("shuffle", False)
        self.chunkseq = 0
        self.bufs = []
        self.n_items = 0
        self.n_tokens = 0
        self.chunk_files = []

    def save_config(self):
        self.config.update(dict(
            n_items=self.n_items,
            n_tokens = self.n_tokens,
            chunkseq=len(self.chunk_files),
            n_chunks=self.n_chunks,
            file_ext=self.file_ext,
            shuffle=self.shuffle,
        ))
        file_checks = make_chunk_filelist(self.store_path, self.chunk_files)
        if file_checks:
            self.config['files'] = file_checks
        
        config_file = safe_join_path(self.store_path, f'{self.prefix}config.json')
        with open(config_file, "w") as w:
            json.dump(self.config, w)
    
    def save(self, save_config=True):
        if len(self.bufs) > 0:
            chunk_file = chunkseq_to_filename(self.chunkseq, self.prefix, self.file_ext)
            save_chunk_file(self.store_path, chunk_file, self.bufs)
            self.chunk_files.append(chunk_file)
            self.n_items += len(self.bufs)
            if len(self.bufs) == self.n_chunks:
                self.chunkseq += 1
                self.bufs = []
        if save_config:
            self.save_config()
            verbose_print(f'トークン数: {self.n_tokens:,} 件数: {self.n_items:,}')

    def append(self, block: List[int]):
        self.bufs.append(np.array(block, dtype=np.int32))
        self.n_tokens += len(block)
        if len(self.bufs) == self.n_chunks:
            self.save(save_config=False)

    def extend(self, blocks: List[List[int]]):
        for block in blocks:
            self.append(block)
        return []   


def split_to_store(filename, N=-1,
                   desc=None,
                   tokenizer_path=DEFAULT_TOKENIZER, 
                   training_type='',
                   format='simple', 
                   split='train',
                   block_size=None, # DEFAULT_BLOCKSIZE 
                   store_path=None, 
                   verbose=True, histogram=False,
                   split_args={}):
    
    if isinstance(tokenizer_path, str):
        tokenizer = load_tokenizer(tokenizer_path)
    else:
        tokenizer = tokenizer_path

    filebase = get_filebase(filename)
    if store_path is None:
        _, _, tokenizer_name = tokenizer_path.rpartition('/')
        store_path=f'{tokenizer_name}/{filebase}'
        verbose_print(f'Saving To.. {store_path}')

    splitter = new_TextSplitter(tokenizer, training_type,
                                format=format, block_size=block_size, 
                                **split_args)
    
    prefix = f'{(splitter.prefix+split)}_'
    store = DatasetStore(store_path, prefix, block_size, **split_args)

    if desc:
        store.config['desc'] = desc
    store.config['tokenizer_path'] = str(tokenizer.name_or_path)
    store.config['tokenizer'] = record_tokenizer(tokenizer)
    store.config['splitter'] = splitter.about()

    iterator = file_iterator(filename, N=N)
    splitter.split_iter(iterator=iterator, update_fn=store.extend)
    splitter.report(store.config, verbose=verbose)
    store.save()
    print(store.config)
    if histogram:
        make_histogram(tokenizer, store_path, store.chunk_files, verbose=verbose)


def make_histogram(tokenizer, store_path, chunk_files, verbose=True):
    token_ids = list(range(0, tokenizer.vocab_size))
    vocabs = tokenizer.convert_ids_to_tokens(token_ids)
    counts = [0] * tokenizer.vocab_size
    csv_file = f'{store_path.replace("/", "_")}.csv'
    from tqdm import tqdm
    for chunk_file in tqdm(chunk_files):
        chunks = load_chunk_file(store_path, chunk_file)
        for chunk in chunks:
            for token_id in chunk:
                counts[token_id] += 1
    df = pd.DataFrame({'tokens': vocabs, 'counts': counts})
    if verbose:
        print(df['counts'].describe())
    df.to_csv(csv_file)
    verbose_print(f"字句の出現頻度を'{csv_file}'に保存しました。")

def make_local_store(filename, tokenizer, block_size, args):
    filebase = get_filebase(filename)
    store_path = safe_join_path(args['cache_dir'],filebase)
    split_to_store(filename, 
                   tokenizer_path=tokenizer, 
                   for_pretraining=args.get('training_type', ''),
                   format='simple', 
                   split='train',
                   block_size=block_size, 
                   store_path=store_path,
                   verbose=False, histogram=False,
                   kwargs={})
    return str(os.path.abspath(store_path))


from typing import TYPE_CHECKING, Any, Dict, List, NamedTuple, Optional, Sequence, Tuple, Union

import os
import re
import hashlib

import torch
from transformers import AutoTokenizer, T5Tokenizer
from .commons import *

DEFAULT_NL = os.environ.get('PT_NK', '<nL>')
DEFAULT_SEP = os.environ.get('PT_SEP', '<seP>')
DEFAULT_OUTPUT_SEP = os.environ.get('PT_OUTPUT_SEP', '<outpuT>')
DEFAULT_ELLIPSIS = os.environ.get('PT_ELLIPSIS', '<ellipsiS>')

def find_token_id(tokenizer: AutoTokenizer, *token: str)->int:
    ids = tokenizer.convert_tokens_to_ids(token)
    for id in ids:
        if id != tokenizer.unk_token_id:
            return id
    return tokenizer.unk_token_id

def find_newline_token_id(tokenizer: AutoTokenizer):
    return find_token_id(tokenizer, DEFAULT_NL, "<nL>", "<nl>")

def find_sep_token_id(tokenizer: AutoTokenizer):
    return find_token_id(tokenizer, DEFAULT_SEP, "<seP>", "<sep>")

def find_output_sep_token_id(tokenizer: AutoTokenizer):
    return find_token_id(tokenizer, DEFAULT_OUTPUT_SEP, "<outpuT>", "<output>")

def find_ellipsis_token_id(tokenizer: AutoTokenizer):
    return find_token_id(tokenizer, DEFAULT_ELLIPSIS, "<ellipsiS>", "<ellipsis>", "<masK>", "<mask>", "<extra_id_99>")

_EXTRA_IDS = [f'<extra_id_{i}>' for i in range(100)]

def find_extra_ids(tokenizer: AutoTokenizer):
    return tokenizer.convert_tokens_to_ids(_EXTRA_IDS)


def _upper_repl(matchobj):
    #print(matchobj, matchobj.group(0))
    return '<cZ>' + matchobj.group(0).lower()

def _cap_repl(matchobj):
    #print(matchobj, matchobj.group(0))
    return matchobj.group(0)[4:].upper()

_UpperPattern = re.compile('([A-Z][a-z])')
_CapitalizedPattern = re.compile(r'(\<cZ\>[a-z])')

def pre_encode(s):
    if isinstance(s, str):
        s = _UpperPattern.sub(_upper_repl, s)
        return s.replace('\t', '    ').replace('\n', '<nL>')
    if isinstance(s, tuple):
        return tuple(map(pre_encode, s))
    return s

def post_decode(s):
    if isinstance(s, str):
        return _CapitalizedPattern.sub(_cap_repl,s).replace('<nL>', '\n')
    return s


def adapt_tokenizer(tokenizer: AutoTokenizer):

    orig_tokenize = tokenizer.tokenize
    def papertown_tokenize(text: str, pair: Optional[str] = None, add_special_tokens: bool = False) -> List[str]:
        text=pre_encode(text)
        pair=pre_encode(pair)
        return orig_tokenize(text)
    tokenizer.tokenize = papertown_tokenize

    orig_encode_plus = tokenizer.encode_plus
    def papertown_encode_plus(
        text,
        text_pair = None,
        add_special_tokens: bool = True,
        padding = False,
        truncation = None,
        max_length: Optional[int] = None,
        stride: int = 0,
        is_split_into_words: bool = False,
        pad_to_multiple_of: Optional[int] = None,
        return_tensors = None,
        return_token_type_ids: Optional[bool] = None,
        return_attention_mask: Optional[bool] = None,
        return_overflowing_tokens: bool = False,
        return_special_tokens_mask: bool = False,
        return_offsets_mapping: bool = False,
        return_length: bool = False,
        verbose: bool = True,
        **kwargs,
    ):
        text = pre_encode(text)
        text_pair=pre_encode(text_pair)
        return orig_encode_plus(text, 
        text_pair=text_pair,
        add_special_tokens = add_special_tokens,
        padding=padding,
        truncation=truncation,
        max_length=max_length,
        stride=stride,
        is_split_into_words=is_split_into_words,
        pad_to_multiple_of=pad_to_multiple_of,
        return_tensors=return_tensors,
        return_token_type_ids=return_token_type_ids,
        return_attention_mask=return_attention_mask,
        return_overflowing_tokens=return_overflowing_tokens,
        return_special_tokens_mask=return_special_tokens_mask,
        return_offsets_mapping=return_offsets_mapping,
        return_length=return_length,
        verbose=verbose,
        **kwargs,
    )
    tokenizer.encode_plus = papertown_encode_plus

    orig_batch_encode_plus = tokenizer.batch_encode_plus
    def papertown_batch_encode_plus(
        batch_text_or_text_pairs,
        add_special_tokens: bool = True,
        padding = False,
        truncation = None,
        max_length = None,
        stride: int = 0,
        is_split_into_words: bool = False,
        pad_to_multiple_of: Optional[int] = None,
        return_tensors = None,
        return_token_type_ids: Optional[bool] = None,
        return_attention_mask: Optional[bool] = None,
        return_overflowing_tokens: bool = False,
        return_special_tokens_mask: bool = False,
        return_offsets_mapping: bool = False,
        return_length: bool = False,
        verbose: bool = True,
        **kwargs,
    ):
        batch_text_or_text_pairs = [pre_encode(x) for x in batch_text_or_text_pairs]        
        return orig_batch_encode_plus(
            batch_text_or_text_pairs=batch_text_or_text_pairs,
            add_special_tokens=add_special_tokens,
            padding=padding,
            truncation=truncation,
            max_length=max_length,
            stride=stride,
            is_split_into_words=is_split_into_words,
            pad_to_multiple_of=pad_to_multiple_of,
            return_tensors=return_tensors,
            return_token_type_ids=return_token_type_ids,
            return_attention_mask=return_attention_mask,
            return_overflowing_tokens=return_overflowing_tokens,
            return_special_tokens_mask=return_special_tokens_mask,
            return_offsets_mapping=return_offsets_mapping,
            return_length=return_length,
            verbose=verbose,
            **kwargs,
        )
    tokenizer.batch_encode_plus = papertown_batch_encode_plus

    orig_convert_tokens_to_string = tokenizer.convert_tokens_to_string
    def papertown_convert_tokens_to_string(tokens: List[str]):
        s = orig_convert_tokens_to_string(tokens)
        return post_decode(s)
    tokenizer.convert_tokens_to_string = papertown_convert_tokens_to_string

    orig_decode = tokenizer.decode
    def papertown_decode(token_ids, 
                         skip_special_tokens: bool = False,
                         clean_up_tokenization_spaces: bool = None,
                        **kwargs):
        s = orig_decode(token_ids, 
                        skip_special_tokens=skip_special_tokens,
                        clean_up_tokenization_spaces=clean_up_tokenization_spaces,
                        **kwargs)
        return post_decode(s)
    tokenizer.decode = papertown_decode


def get_tokenizer_info(tokenizer: AutoTokenizer):
    allvoc = ''.join(tokenizer.get_vocab().keys())
    sha256 = hashlib.sha256(allvoc.encode()).hexdigest()
    return dict(
        name_or_path=tokenizer.name_or_path,
        pad_token_id = tokenizer.pad_token_id,
        eos_token_id = tokenizer.eos_token_id,
        hash=sha256, 
        vocab_size=tokenizer.vocab_size)


def load_tokenizer(tokenizer_path=DEFAULT_TOKENIZER, adapt=True):
    tokenizer = AutoTokenizer.from_pretrained(tokenizer_path, legacy=False, trust_remote_code=True, use_fast=False)
    if adapt:
        newline_token_id = find_token_id(tokenizer, "<nL>")
        if newline_token_id != tokenizer.unk_token_id:
            adapt_tokenizer(tokenizer)
    return tokenizer





import os
import argparse
#from papertown import DatasetStore, DataComposer, load_tokenizer
#from tqdm import tqdm

from .commons import *
from .splitters import split_to_store

def _tobool(s):
    return s.lower() == 'true' or s == '1'

def _is_float(s):
    try:
        float(s)
        return True
    except ValueError:
        return False

def _parse_args(args):
    args_list = args.split(',')
    args_dict = {}
    for arg in args_list:
        key, value = arg.split('=')
        if value.isdigit():
            args_dict[key] = int(value)
        elif _is_float(value):
            args_dict[key] = float(value)
        elif value.lower() == 'true' or value.lower() == 'false':
            args_dict[key] = _tobool(value)
        else:
            args_dict[key] = value
    return args_dict

def setup_store():
    parser = argparse.ArgumentParser(description="papertown_store")
    parser.add_argument("files", type=str, nargs="+", help="files")
    parser.add_argument("--desc", type=str, default=None)
    parser.add_argument("--tokenizer_path", default=DEFAULT_TOKENIZER)
    parser.add_argument("--store_path", default=None)
    parser.add_argument("--block_size", type=int, default=None)
    parser.add_argument("--type", type=str, default='')
    parser.add_argument("--format", default="simple")
    parser.add_argument("--split", default="train")
    parser.add_argument("--split_args", type=_parse_args, default=None)
    parser.add_argument("--sep", type=str, default=None)
    parser.add_argument("--output", type=str, default=None)
    parser.add_argument("--N", "-N", type=int, default=-1)
    parser.add_argument("--verbose", type=_tobool, default=True)
    parser.add_argument("--histogram", type=_tobool, default=False)
    parser.add_argument("--num_works", type=int, default=0)
    hparams = parser.parse_args()  # hparams になる
    return hparams

def main_store():
    hparams = setup_store()
    split_to_store(
        hparams.files[0],
        N=hparams.N,
        desc=hparams.desc,
        tokenizer_path=hparams.tokenizer_path,
        training_type=hparams.type,
        format=hparams.format,
        split=hparams.split,
        split_args=hparams.split_args or {},
        block_size=hparams.block_size, 
        store_path=None, 
        verbose=True, 
        histogram=hparams.histogram
    )


def main_update():
    import os
    os.system('pip3 uninstall -y papertown')
    os.system('pip3 install -U git+https://github.com/kuramitsulab/papertown.git')



from typing import Any, List

import os
import random

import json
import shutil
from urllib.parse import urlparse, parse_qs
import hashlib

from collections import deque

from filelock import FileLock
import numpy as np

import torch
from torch.utils.data import Dataset
from transformers import DataCollatorForLanguageModeling

from .commons import *
from .file_utils import *
from .tokenizers import *
from .splitters import *

_ID = 0
def random_name():
    global _ID
    _ID+= 1
    return f'Cache{_ID-1}'

# 設定ファイル

DEFAULT_BLOCK_SIZE = 2096
DEFAULT_MAX_LENGTH = 4096
N_CHUNKS = 4096

# ChunkedDataset

class _DummyFileLock:
    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        pass

def _FileLock(lockfile: str):
    # lockがNoneなら何もしない
    return _DummyFileLock() if lockfile is None else FileLock(lockfile)


def get_rank():
    if torch.distributed.is_available() and torch.distributed.is_initialized():
        return torch.distributed.get_rank()
    else:
        return 0

def get_world_size():
    if torch.distributed.is_available() and torch.distributed.is_initialized():
        return torch.distributed.get_world_size()
    else:
        return 1

def url_to_hash(url):
    return hashlib.md5(url.encode()).hexdigest()


class ChunkedDataset(Dataset):
    def __init__(self, url:str, url_args: dict, split='train', block_size=None, prefetch=1):
        """
        block_size=Noneのときは、再分割しない (ファインチューニングは再分割しない)
        """
        self.url = safe_dir(url)
        self.url_args = url_args
        self.split = split
        self.block_size = block_size
        self.cache_dir = safe_join_path(url_args['cache_dir'], url_to_hash(url))
        self.lock_file = url_args['lock_file']
        if url_args['training_type'].startswith('pre'):
            self.split_prefix = f"pretrain_"
        else:
            self.split_prefix = f"{split}_"
        self.load_config()
        self.queue = deque(maxlen=64)
        self.cache = {}
        self.prefetch=1
        if self.prefetch > 0 and self.n_items > 0:
            self.try_prefetch(0)

    def get_valid_dataset(self, split='valid'):
        dataset = ChunkedDataset(self.url, self.url_args, split=split, block_size=self.block_size)
        if len(dataset) > 0:
            return dataset
        return None

    def load_config(self):
        with _FileLock(self.lock_file):
            config_file = resolve_file(self.url, f'{self.split_prefix}config.json', self.cache_dir)
        try:
            with open(config_file) as f:
                config = json.load(f)
        except BaseException as e:
            verbose_print(f'見つかりません {self.url} ({config_file})')
            config = dict(n_items=0, n_tokens=0)
        # 設定
        self.file_ext = config.get("file_ext", "npz")
        self.n_tokens = config.get('n_tokens', 0)
        self.n_items = config.get("n_items", 0)
        self.n_chunks = config.get("n_chunks", N_CHUNKS)
        self.tokenizer_path = config.get("tokenizer_path", DEFAULT_TOKENIZER)
        self.max_chunkseq = max(config.get("chunkseq", 1), 1)  # 古いデータ用
        self.n_subblocks = 1
        if self.block_size is not None and 'block_size' in config and self.block_size < config['block_size']:
            self.n_subblocks = config['block_size'] // self.block_size
            verbose_print(f'{self.url} を{self.n_subblocks}個に再分割します')
        self.is_seq2seq = 'output_sep_token_id' in config
        self.config = config
        return config

    def __len__(self):
        return self.n_items * self.n_subblocks

    def try_prefetch(self, chunkseq):
        chunk_file = chunkseq_to_filename(chunkseq % self.max_chunkseq, self.split_prefix, self.file_ext)
        resolve_file(self.url, chunk_file, self.cache_dir, sync=False)

    def get_chunks(self, chunk_file):
        if chunk_file in self.cache:
            return self.cache[chunk_file]
        with _FileLock(self.lock_file):
            chunk_file2 = resolve_file(self.url, chunk_file, self.cache_dir)
            chunks = load_chunk_file(chunk_file2)
            chunks = self.resize_chunks(chunks)
        if chunks is None:
            # エラーで落ちるくらいなら、キャッシュのデータで学習を続ける
            chunks = self.cache[self.queue[0]]
        # if self.shuffle:
        #     random.shuffle(chunks)
        if len(self.queue) == 64:
            older = self.queue.popleft()
            if older in self.cache:
                del self.cache[older]
        self.queue.append(chunk_file)
        self.cache[chunk_file] = chunks
        return chunks

    def resize_chunks(self, chunks):
        if self.n_subblocks > 1:
            newchunks=[]
            for chunk in chunks:
                splits = np.array_split(chunk, self.n_subblocks)
                newchunks.extend(splits)
            return newchunks
        return chunks

    def __getitem__(self, index):
        i = index
        chunkseq = i // self.n_chunks
        chunk_file = chunkseq_to_filename(chunkseq, self.split_prefix, self.file_ext)
        chunks = self.get_chunks(chunk_file)
        if self.prefetch > 0 and index % self.n_chunks == 0:
            self.try_prefetch(chunkseq+self.prefetch)
        return chunks[i % self.n_chunks]
    

class DistributedIndexer(Dataset):
    def __init__(self, dataset, url_args:dict):
        self.dataset = dataset
        self.url_args = url_args
        self.count = 0
        self.dataset_size = len(dataset)
        self.valid_dataset = None

        # offset をパラメータから調整する
        self.offset = url_args.get('start', 0)
        if isinstance(self.offset, float) and self.offset < 1.0:
            self.offset = int(self.offset * self.dataset_size)
        self.offset = self.offset % self.dataset_size
        
        # length をパラメータから調整する
        self.length = url_args.get('length', self.dataset_size)
        if isinstance(self.length, float) and self.length < 1.0:
            self.length = int(self.length * self.dataset_size)
        if self.length > self.dataset_size:
            self.length = self.length % self.dataset_size
        self.epoch = 1
        # DistributedSamplerと同様に均等に分割して各プロセスが同じデータを読まないようにする
        self.sublength = self.length // get_world_size()
        if get_world_size() > 1:
            verbose_print('ランクごとに再配置します')
            self.offset = self.offset + (get_rank() * self.sublength)

    def __len__(self):
        return self.length

    def __getitem__(self, idx):
        index = (self.offset + self.count) % self.dataset_size
        self.count += 1
        if self.count >= self.sublength:
            self.count = 0
            self.epoch += 1
        return self.dataset[index]

    def get_valid_dataset(self, valid_split=0.1):
        if self.valid_dataset:
            return self.valid_dataset
        self.valid_dataset = self.dataset.get_valid_dataset()
        if self.valid_dataset:
            return self.valid_dataset
        # 訓練データを 9:1 に分割して検証データを作る
        valid_dataset = DistributedIndexer(self.dataset, self.url_args)
        valid_dataset.length = int(self.length * valid_split)
        self.length -= valid_dataset.length
        sublength = valid_dataset.sublength
        valid_dataset.sublength = int(self.sublength * valid_split)
        self.sublength -= valid_dataset.sublength
        valid_dataset.offset += (sublength - valid_dataset.sublength)
        self.valid_dataset = valid_dataset
        return self.valid_dataset

    # def get_num_of_tokens(self):
    #     if len(self.dataset) == self.length:
    #         return self.n_tokens
    #     return self.n_tokens * self.length // len(self.dataset)


def build_inputs_for_clm(data, max_length):
    return torch.tensor(data[:max_length].astype(np.int64), dtype=torch.long)

def parse_url_list(url_list=[]):
    if isinstance(url_list, str):
        if os.path.exists(url_list):
            with open(url_list) as f:
                return [url.strip() for url in f.readlines() if url.strip() != '' and not url.startswith('#')]
        return url_list.split('|')
    return url_list


def convert_to_number(value):
    """
    文字列を可能ならば整数または浮動小数点数に変換する。
    変換できない場合はそのままの文字列を返す。
    """
    lower_string = str(value).lower()
    if lower_string == 'true':
        return True
    if lower_string == 'false':
        return False
    try:
        return int(value)
    except ValueError:
        try:
            return float(value)
        except ValueError:
            return str(value)

def _recalculate_length(mixed):
    dd={}
    n_items = 0
    for ds in mixed:
        if id(ds) not in dd:
            dd[id(ds)] = ds
            n_items += len(ds)
    return n_items

class MixingDataset(Dataset):
    def __init__(self, mixed, n_items, build_fn, max_length):
        self.mixed = mixed
        self.mixing = len(self.mixed)
        self.n_items = n_items
        self.build_fn = build_fn
        self.max_length = max_length

    def __len__(self):
        return self.n_items

    def __getitem__(self, index):
        data = self.mixed[index % self.mixing][index]
        return self.build_fn(data, self.max_length)

    def get_valid_dataset(self, valid_split=0.1):
        mixed = []
        for ds in self.mixed:
            ds = ds.get_valid_dataset(valid_split)
            mixed.append(ds)
        train_size = _recalculate_length(self.mixed)
        valid_size = _recalculate_length(mixed)
        verbose_error(f'訓練データ {train_size} 検証データ {valid_size}')
        self.n_items = min(train_size, self.n_items)
        return MixingDataset(mixed, min(valid_size, self.n_items // 4), self.build_fn, self.max_length)


class DataComposer(MixingDataset):
    def __init__(self, 
                 url_list, training_type="pre", split = DEFAULT_SPLIT, 
                 block_size=None, max_length = DEFAULT_MAX_LENGTH,
                 build_fn=build_inputs_for_clm, 
                 cache_dir = DEFAULT_CACHE_DIR, 
                 random_seed=None, shuffle=True,
                 tokenizer=None, 
                 use_filelock=True, cleanup=True,
                 prefetch=1, test_run=None):
        self.training_type = training_type
        self.split = split
        self.block_size=block_size
        self.max_length = max_length
        self.cache_dir = f'{safe_dir(cache_dir)}/{random_name()}'
        os.makedirs(self.cache_dir, exist_ok=True)
        self.lock_file = f'{self.cache_dir}/lock.me' if use_filelock else None
        self.random_seed=getint_from_environ('PT_RANDOM_SEED', random_seed, 42)
        self.tokenizer_path = None
        self.prepare_data(parse_url_list(url_list), block_size, tokenizer)
        self.cleanup = False if get_rank() > 0 else cleanup
        self.prefetch = prefetch
        self.build_fn = build_fn
        # テスト実行
        test_run = getint_from_environ('PT_TEST_RUN', test_run, None)
        if test_run and isinstance(test_run, int):
            verbose_print('反復を減らして、テスト実行します', test_run)
            self.n_items = min(test_run, self.n_items)

    def parse_url_args(self, url):
        parsed_url = urlparse(url)
        if len(parsed_url.scheme):
            base_url = f"{parsed_url.scheme}://{parsed_url.netloc}{parsed_url.path}"
        else:
            base_url = f"{parsed_url.path}"
        params = parse_qs(parsed_url.query)
        url_args = {k: convert_to_number(v[0]) for k, v in params.items()}
        url_args['training_type'] = self.training_type
        if 'split' not in url_args:  # 個別のsplitの指定がなければ... 
            url_args['split'] = self.split
        url_args['prefix'] = f'{self.training_type}{url_args["split"]}_'
        url_args['cache_dir'] = self.cache_dir
        url_args['lock_file'] = self.lock_file
        return base_url, url_args

    def prepare_data(self, urls, block_size, tokenizer=None):
        self.n_items = 0
        self.n_tokens = 0
        datasets = []
        for url in urls:
            url, url_args = self.parse_url_args(url)
            if url.endswith('.gz') or url.endswith('.jsonl') or url.endswith('.txt'):
                tokenizer = self.prepare_tokenizer(tokenizer)
                url = make_local_store(url, tokenizer, block_size, url_args)
            dataset = ChunkedDataset(url, url_args, split=self.split, block_size=block_size)
            if self.check_tokenizer(url, dataset) == False:
                continue
            if self.training_type.startswith('seq2') and not dataset.is_seq2seq:
                verbose_print(f'** {url} は、seq2seqに対応していません。無視して学習を続けます。')
                continue
            if len(dataset) == 0:
                verbose_print(f'** {url} は、スキップして学習を続けます。')
                continue
            dataset = DistributedIndexer(dataset, url_args)
            datasets.append(dataset)
            self.n_items += len(dataset)
            # n_tokens = ds.get_num_of_tokens()
            # verbose_print(f'{url} トークン数: {format_unit(n_tokens)} {n_tokens:,} 件数: {len(ds):,}')
        if self.n_items > 0:
            self.blend_data(datasets)

    def blend_data(self, datasets: List[DistributedIndexer]):
        lens = [len(ds) for ds in datasets]
        total = sum(lens)
        mixer_base = (total // min(lens))+1
        lens = [int((dlen * mixer_base) / total) for dlen in lens]
        verbose_print('ミキサーパターン:', lens)
        self.mixed = []
        for dlen, ds in zip(lens, datasets):
            self.mixed.extend([ds]*dlen)
        random.seed(self.random_seed)
        random.shuffle(self.mixed)
        self.mixing = len(self.mixed)

    def prepare_tokenizer(self, tokenizer=None):
        if tokenizer is not None:
            return tokenizer
        if self.tokenizer_path is None:
            verbose_print(f'トークンナイザーの指定がないので DEFAULT_TOKENIZER={DEFAULT_TOKENIZER}を使います')
            self.tokenizer_path = DEFAULT_TOKENIZER
        return load_tokenizer(self.tokenizer_path)

    def check_tokenizer(self, url, dataset):
        if self.tokenizer_path is None:
            self.tokenizer_path = dataset.tokenizer_path
        elif self.tokenizer_path != dataset.tokenizer_path:
            verbose_error(f'警告: トークンナイザーが一致しません。{self.tokenizer_path}')
            verbose_error(f'    {dataset.tokenizer_path} @{url}')
            verbose_print(f'** {url} は、スキップして学習を続けます。')
            return False
        return True

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.n_items = 0
        self.mixed = None
        if self.cleanup and os.path.isdir(self.cache_dir):
            try:
                shutil.rmtree(self.cache_dir)
                verbose_print('Cleaned up', self.cache_dir)
            except:
                pass

    # def __len__(self):
    #     return self.n_items

    # def __getitem__(self, idx):
    #     mix = len(self.mixer)
    #     item = self.mixer[idx % mix][idx]
    #     return self.build_fn(item, self.max_length)




class PretrainComposer(DataComposer):
    def __init__(self, url_list, block_size, **kwargs):
        kwargs['training_type'] = 'pre'
        kwargs['split'] = 'train'
        kwargs['max_length'] = block_size
        DataComposer.__init__(self, url_list, block_size=block_size, **kwargs)

    def get_collator(self):
        tokenizer = load_tokenizer(self.tokenizer_path)
        return DataCollatorForLanguageModeling(tokenizer, mlm=False)


class FinetuneComposer(DataComposer):
    def __init__(self, url_list, split="train", **kwargs):
        kwargs['training_type'] = ''
        if 'block_size' in kwargs: # ブロックサイズは、事前学習用のパラメータ
            kwargs['max_length'] = kwargs['block_size']
            del kwargs['block_size']
        DataComposer.__init__(self, url_list, split=split, **kwargs)

    def get_collator(self):
        tokenizer = load_tokenizer(self.tokenizer_path)
        return DataCollatorForLanguageModeling(tokenizer, mlm=False)


class DP(object):
    def __init__(self, tokenizer, lambda_=20):
        self.lambda_ = lambda_
        self.eos_token_id = tokenizer.eos_token_id
        self.extra_ids = find_extra_ids(tokenizer)
        self.newline_id = find_newline_token_id(tokenizer)

    def __call__(self, data, max_length):
        index = 0
        start = 0
        size = min(max_length, len(data))
        for i, length in enumerate(np.random.poisson(self.lambda_, 1000), start=1):
            start = start + max(1, length) * i
            if start >= size:
                break
            if data[start] != self.eos_token_id or data[start] != self.newline_id:
                data[start] = self.extra_ids[index]
                index+=1
        return torch.tensor(data[:max_length].astype(np.int64), dtype=torch.long)




import gc
import torch
torch.backends.cuda.matmul.allow_tf32=True
from transformers import DataCollatorForLanguageModeling
from transformers import Trainer, TrainingArguments

import numpy as np
from datasets import Dataset

from .commons import *

def count_parameters(model)->int:
    """
    モデルのパラメータ数を数える

    model: モデル
    return パラメータ数
    """
    return sum(p.numel() for p in model.parameters())


def print_model(model):
    n_parameters=count_parameters(model)
    config = model.config
    print(f'Parameters: {n_parameters} {format_unit(n_parameters)}', end=' ')
    if hasattr(config, 'max_position_embeddings'):
        print(f"max_length: {config.max_position_embeddings}", end=' ')
    elif hasattr(config, "n_positions"):
        print(f"max_length: {config.n_positions}", end=' ')
    print(f"vocab_size: {config.vocab_size}")

    if hasattr(config, 'd_kv'):  # T5
        print(f"d_model: {model.config.d_model}", end=' ')
        print(f"d_kv: {model.config.d_kv}", end=' ')
        print(f"d_ff: {model.config.d_ff}", end=' ')
        print(f"num_heads: {model.config.num_heads}", end=' ')
        print(f"num_layers: {model.config.num_layers}+{model.config.num_decoder_layers}")
        print(config)
    elif hasattr(config, 'n_embd'): #GPT-2
        print(f"hidden_size: {config.n_embd}", end=' ')
        print(f"intermediate_size: {config.n_inner}", end=' ')
        print(f"n_dims: {config.n_embd//config.n_head}", end=' ')
        print(f"n_heads: {config.n_head}", end=' ')
        print(f"n_layers: {config.n_layer}")
        print(config)
    elif hasattr(config, 'hidden_size'): #GPT-NeoX
        print(f"hidden_size: {config.hidden_size}", end=' ')
        print(f"intermediate_size: {config.intermediate_size}", end=' ')
        print(f"n_dims: {config.hidden_size//model.config.num_attention_heads}", end=' ')
        print(f"n_heads: {config.num_attention_heads}", end=' ')
        print(f"n_layers: {config.num_hidden_layers}")
    else:
        print(config)

def print_gpu_utilization():
    try:
        from pynvml import nvmlInit,nvmlDeviceGetHandleByIndex,nvmlDeviceGetMemoryInfo
        nvmlInit()
        handle = nvmlDeviceGetHandleByIndex(0)
        info = nvmlDeviceGetMemoryInfo(handle)
        print(f"GPU memory occupied: {format_unit(info.used, scale=1024)}iB.")
    except:
        pass

def dummy_dataset(max_length, dataset_size=1024):
    dummy_data = {
        "input_ids": (np.arange(100, dataset_size*max_length+100) % 15000).reshape((dataset_size, max_length))
    }
    ds = Dataset.from_dict(dummy_data)
    ds.set_format("pt")
    return ds

def print_summary(result, use_flash=False):
    m = result.metrics
    print(f"Time: {m['train_runtime']:.2f}  {format_unit(m['train_runtime'], scale=60)}", end=' ')
    print(f"Samples/second: {m['train_samples_per_second']:.2f} FlashAttn={use_flash}")
    print(f"Global step: {result.global_step} batch_size: {1024//result.global_step}", end=' ')
    if 'total_flos' in m:
        print(f"FLOS: {m['total_flos']} {format_unit(m['total_flos'])} Loss: {m['train_loss']:.5f}")
    else:
        print(f"Loss: {m['train_loss']:.5f}")
    print_gpu_utilization()


def train_model(model, tokenizer, max_length, use_fp16=False, use_flash=False):

    if use_flash:
        from optimum.bettertransformer import BetterTransformer
        model = BetterTransformer.transform(model)

    data_collator = DataCollatorForLanguageModeling(tokenizer, mlm=False)

    args = TrainingArguments(
        output_dir="./output",
        overwrite_output_dir=True,
        per_device_train_batch_size=128,
        per_device_eval_batch_size=128,
        auto_find_batch_size=True,  # バッチサイズ自動
        do_eval=False,
        logging_steps=1000,
#        gradient_accumulation_steps=1,
        num_train_epochs=1,
        weight_decay=0.1,
#        warmup_steps=1_000,
        lr_scheduler_type="cosine",
        learning_rate=5e-4, #TODO: 論文から探す
#        save_steps=5_000,
        fp16=use_fp16,
    )

    trainer = Trainer(
        model=model,
        args=args,
        train_dataset=dummy_dataset(max_length),
        data_collator=data_collator,
    )
    result = trainer.train()
    print_summary(result, use_flash)
    # モデルを保存 output_path に保存します
    if use_flash:
        model = BetterTransformer.reverse(model)
    output_path='trained'
    tokenizer.save_pretrained(output_path)
    model.save_pretrained(output_path)
    model.cpu()
    gc.collect()
    torch.cuda.empty_cache()
    print_gpu_utilization()


def new_T5(d_model=256, d_kv=32, d_ff=1024, n_head=6, n_layers=12, max_length=2048, tokenizer=DEFAULT_TOKENIZER):
    from transformers import T5Config, T5ForConditionalGeneration, AutoTokenizer

    if isinstance(tokenizer, str):
        tokenizer = AutoTokenizer.from_pretrained(tokenizer, legacy=False, trust_remote_code=True, use_fast=False)

    config = T5Config(
        vocab_size = len(tokenizer),
        d_model = d_model,
        d_kv = d_kv,
        d_ff = d_ff,
        num_layers = n_layers,
        num_decoder_layers = None,
        num_heads = n_head,
        relative_attention_num_buckets = 32,
        relative_attention_max_distance = 128,
        dropout_rate = 0.1,
        layer_norm_epsilon = 1e-06,
        initializer_factor = 1.0,
        feed_forward_proj = 'gated-gelu',
        is_encoder_decoder = True,
        use_cache = True,
        tokenizer_class = 'T5Tokenizer',
        tie_word_embeddings = False,
        pad_token_id = tokenizer.pad_token_id,
        eos_token_id = tokenizer.eos_token_id,
        decoder_start_token_id=0,
    )

    model = T5ForConditionalGeneration(config)
    print_model(model)
    return model


# GPT-2

def new_GPT2(max_length=2048, n_dims=64, n_heads=6, n_layers=12, intermediate_size=1024, tokenizer=DEFAULT_TOKENIZER):
    from transformers import AutoTokenizer, GPT2LMHeadModel, GPT2Config

    if isinstance(tokenizer, str):
        tokenizer = AutoTokenizer.from_pretrained(tokenizer, legacy=False, trust_remote_code=True, use_fast=False)

    config = GPT2Config(
        vocab_size = len(tokenizer),
        bos_token_id = tokenizer.eos_token_id,
        eos_token_id = tokenizer.eos_token_id,
        pad_token_id = tokenizer.pad_token_id,
        n_positions=max_length,
        n_ctx=max_length,
        n_embd=n_dims*n_heads,
        n_head=n_heads,
        n_layer=n_layers,
        n_inner=intermediate_size,
    )

    model = GPT2LMHeadModel(config)
    print_model(model)
    return model

# GPTNeoX

def new_GPTNeoX(max_length=2048, n_dims=64, n_heads=12, n_layers=12, intermediate_size=2048, tokenizer=DEFAULT_TOKENIZER):
    from transformers import AutoTokenizer, GPTNeoXForCausalLM, GPTNeoXConfig
    if isinstance(tokenizer, str):
        tokenizer = AutoTokenizer.from_pretrained(tokenizer, legacy=False, trust_remote_code=True, use_fast=False)

    config = GPTNeoXConfig(
        vocab_size = len(tokenizer),
        pad_token_id = tokenizer.pad_token_id,
        bos_token_id = tokenizer.eos_token_id,
        eos_token_id = tokenizer.eos_token_id,
        max_position_embeddings=max_length, #トークン数
        hidden_size=n_dims * n_heads,
        num_attention_heads = n_heads, #8
        num_hidden_layers = n_layers, #28
        intermediate_size=intermediate_size,
    )

    model = GPTNeoXForCausalLM(config)
    print_model(model)
    return model


## new_Lamma2

def new_Llama2(max_length=2048, n_dims=128, n_heads=8, n_layers=28, intermediate_size=4096, tokenizer=DEFAULT_TOKENIZER):
    from transformers import AutoTokenizer, LlamaForCausalLM, LlamaConfig

    if isinstance(tokenizer, str):
        tokenizer = AutoTokenizer.from_pretrained(tokenizer, legacy=False, trust_remote_code=True, use_fast=False)

    config = LlamaConfig(
        vocab_size = len(tokenizer),
        pad_token_id = tokenizer.pad_token_id,
        bos_token_id = tokenizer.eos_token_id,
        eos_token_id = tokenizer.eos_token_id,
        max_position_embeddings=max_length, #トークン数
        hidden_size=n_dims * n_heads,
        num_attention_heads = n_heads, #8
        num_hidden_layers = n_layers, #28
        intermediate_size=intermediate_size,
    )

    model = LlamaForCausalLM(config)
    print_model(model)
    return model






from typing import Optional
import os

def safe_dir(dir):
    if dir.endswith('/'):
        dir = dir[:-1]
    return dir

def safe_join(dir, file):
    if dir.endswith('/'):
        dir = dir[:-1]
    if file.startswith('/'):
        file = file[1:]
    return f'{dir}/{file}'

DEFAULT_TOKENIZER = os.environ.get('PT_TOKENIZER', 'kkuramitsu/kawagoe')
DEFAULT_SPLIT='train'
DEFAULT_CACHE_DIR = safe_dir(os.environ.get('PT_CACHE_DIR', '.'))

DEFAULT_BLOCK_SIZE = 2096
DEFAULT_MAX_LENGTH = 4096
N_CHUNKS = 4096


def getint(kwargs:dict, 
           key:str, default_value=0, 
           specified:Optional[int]=None) -> int:
    if isinstance(specified, int):
        return specified
    try:
        return int(kwargs.get(key, default_value))
    except:
        return default_value

def getfloat(kwargs:dict, 
             key:str, default_value=0.0,
             specified:Optional[float]=None) -> float:
    if specified:
        return float(specified)
    try:
        return float(kwargs.get(key, default_value))
    except:
        return default_value

def getint_from_environ(key, given=None, default=None):
    return getint(os.environ, key, 
                  default_value=default, 
                  specified=given)


def format_unit(num: int, scale=1000)->str:
    """
    大きな数をSI単位系に変換して返す
    """
    if scale == 1024:
        if num < scale:
            return str(num)
        elif num < scale**2:
            return f"{num / scale:.1f}K"
        elif num < scale**3:
            return f"{num / scale**2:.1f}M"
        elif num < scale**4:
            return f"{num / scale**3:.1f}G"
        elif num < scale**5:
            return f"{num / scale**4:.1f}T"
        elif num < scale**6:
            return f"{num / scale**5:.1f}P"
        else:
            return f"{num / scale**6:.1f}Exa"
    elif scale == 60:
        if num < 1.0:
            return f"{num * 1000:.1f}ms"
        if num < scale:
            return f"{num:.1f}sec"
        elif num < scale**2:
            return f"{num / scale:.1f}min"
        elif num < (scale**2)*24:
            return f"{num /(scale**2):.1f}h"
        else:
            num2 = num % (scale**2)*24
            return f"{num//(scale**2)*24}d {num2/(scale**2):.1f}h"
    else:
        if num < 1_000:
            return str(num)
        elif num < 1_000_000:
            return f"{num / 1_000:.1f}K"
        elif num < 1_000_000_000:
            return f"{num / 1_000_000:.1f}M"
        elif num < 1_000_000_000_000:
            return f"{num / 1_000_000_000:.1f}B"
        else:
            return f"{num / 1_000_000_000_000:.1f}T"


def verbose_print(*args, **kwargs):
    """
    PaperTown 用のデバッグプリント
    """
    print('🏙', *args, **kwargs)

def verbose_error(*args, **kwargs):
    """
    PaperTownのエラープリント
    """
    print('🌆', *args, **kwargs)



import os
import time
import random
from pathlib import Path

import json
import hashlib
import subprocess
from typing import List

# from collections import deque
# from filelock import FileLock

import numpy as np
import gzip
import pyzstd

from tqdm import tqdm

#import torch
from transformers import AutoTokenizer
from torch.utils.data import Dataset

from .commons import *

_ID = 0

def random_name():
    global _ID
    _ID+= 1
    return f'Cache{_ID-1}'

# ファイルシステム

def safe_dir(dir):
    if dir.endswith('/'):
        dir = dir[:-1]
    return dir

def safe_join_path(dir, file):
    if file is None: 
        return dir
    if dir.endswith('/'):
        dir = dir[:-1]
    if file.startswith('/'):
        file = file[1:]
    return f'{dir}/{file}'

def get_filebase(filename):
    filebase = filename
    if '/' in filebase:
        _, _, filebase = filebase.rpartition('/')
    filebase, _, _ = filebase.partition('.')
    return filebase

## file 

def zopen(filepath):
    if filepath.endswith('.gz'):
        return gzip.open(filepath, 'rt')
    elif filepath.endswith('.zstd'):
        return pyzstd.open(filepath, 'rt')
    else:
        return open(filepath, 'r')

def get_filelines(filepath):
    with zopen(filepath) as f:
        line = f.readline()
        c=1
        while line:
            line = f.readline()
            c+=1
    return c

def parse_strip(s):
    return s.strip().replace('<nL>', '\n')

def parse_jsonl(line):
    d = json.loads(line)
    if 'out' in d:
        return f"{d['in']}<outpuT>{d['out']}"
    return d['text']

def file_iterator(filename, N=None, kwargs={}):
    if N == -1:
        N = get_filelines(filename)-1
    if N:
        from tqdm import tqdm
        pbar = tqdm(total=N, desc=filename)
    parse_fn = parse_strip
    if '.json' in filename:
        parse_fn = parse_jsonl
    c=0
    with zopen(filename) as f:
        line = f.readline()
        while line is not None:
            c+=1
            if N: 
                pbar.update()
                if c > N: break
            yield parse_fn(line)
            line = f.readline()
    if N:
        pbar.close()



def _remove_heading_nL(s):
    while s.startswith('<nL>'):
        s = s[4:]
    return s

def _makedirs(path):
    dir, _,  file = path.rpartition("/")
    if '.' in file: #拡張子が含まれる場合
        os.makedirs(dir,  exist_ok=True)
    elif not os.path.isfile(path):
        os.makedirs(path,  exist_ok=True)

def get_file_sha1(filepath: str):
    # ファイルをバイナリモードで読み込む
    with open(filepath, 'rb') as f:
        # ファイルの内容を読み込む
        content = f.read()
        # SHA-1ハッシュオブジェクトを作成
        sha1 = hashlib.sha1()
        # ファイルの内容をハッシュオブジェクトに追加
        sha1.update(content)
        # 16進数でハッシュ値を取得
        sha1_hexdigest = sha1.hexdigest()
    return sha1_hexdigest

def get_filesize(file_path):
    if os.path.exists(file_path) and os.path.isfile(file_path):
        return os.path.getsize(file_path)
    else:
        return -1

def touch(file_path):
    file = Path(file_path)
    file.touch(exist_ok=True)

def wait_for_file(file_path, timeout=60):
    """
    指定されたファイルが存在するかを定期的にチェックし、
    タイムアウトまでにファイルが見つかった場合は True を返します。
    タイムアウトした場合は False を返します。
    """
    start_time = time.time()
    end_time = start_time + timeout
    while time.time() < end_time:
        if get_filesize(file_path) > 0:
            verbose_print(f'{time.time()-start_time} 秒, 待ちました')
            return True  # ファイルが見つかった
        time.sleep(1)  # 1秒待つ
    return False  # タイムアウト

def resolve_file(url_base, file_path, cache_dir, sync=True):
    remote_file = safe_join_path(url_base, file_path)
    if remote_file.startswith('/'):
        # ローカルなファイルパスの場合
        return remote_file
    cache_file = safe_join_path(cache_dir, file_path)
    # ディレクトリを作っておく
    os.makedirs(cache_file.rpartition("/")[0], exist_ok=True)
    cache_file_size = get_filesize(cache_file)
    #print('@', cache_file_size, cache_file)
    if cache_file_size > 0:
        return cache_file

    # ダウンロードコマンド
    if remote_file.startswith('file:'):
        remote_file = os.path.abspath(remote_file[5:]) # file: をとる
        cmd = f'cp {remote_file} {cache_file}'
    else:
        cmd = f"wget -qO {cache_file}.tmp {remote_file} && mv {cache_file}.tmp {cache_file}"

    if sync:
        if cache_file_size == 0:
            verbose_print('ダウンロード中 最大30秒待ちます.', remote_file)
            if wait_for_file(cache_file, 30):
                return cache_file
        touch(cache_file)
        subprocess.call(cmd, shell=True)
        verbose_print(f'Downloaded {get_filesize(cache_file):,} bytes:', cmd)
        return cache_file

    if get_filesize(cache_file) == -1:
        touch(cache_file)
        verbose_print('プレフェッチ', remote_file)
        subprocess.call(f"{cmd} &", shell=True, stderr=subprocess.DEVNULL)
    # else:
    #     verbose_print('既にダウンロード中..', remote_file)
    return None


# chunk file 

def chunkseq_to_filename(chunkseq:int, prefix:str, file_ext:str):
    dir = f"{(chunkseq//100):04d}"
    return f"{dir}/{prefix}{(chunkseq%100):02d}.{file_ext}"

def save_chunk_file(base_dir:str, chunk_file:str, chunks:List[np.ndarray]):
    filepath = safe_join_path(base_dir, chunk_file)
    _makedirs(filepath)
    if filepath.endswith('.npz'):
        np.savez_compressed(filepath, *chunks)
    # return {'filesize': get_filesize(filepath), 
    #         'sha1': get_file_sha1(filepath)}

def load_chunk_file(base_dir:str, chunk_file:str=None):
    filepath = safe_join_path(base_dir, chunk_file)
    try:
        #if filepath.endswith('.npz'):
        npz = np.load(filepath)
        return [npz[n] for n in npz.files]
    except BaseException as e:
        verbose_print(f'broken chunk file {chunk_file}: {e}')
        return None

def check_chunk_file(base_dir:str, chunk_file:str, checks: dict):
    filepath = safe_join_path(base_dir, chunk_file)
    if 'filesize' in checks:
        if get_filesize(filepath) != checks['filesize']:
            return False
    if 'sha1' in checks:
        if get_file_sha1(filepath) != checks['sha1']:
            return False
    return True

def make_chunk_filelist(base_dir:str, chunk_files:List[str]):
    d = {}
    for chunk_file in tqdm(chunk_files, desc='file validation..'):
        if not load_chunk_file(base_dir, chunk_file):
            return None
        filepath = safe_join_path(base_dir, chunk_file)
        checks = {'filesize': get_filesize(filepath), 'sha1': get_file_sha1(filepath)}
        if not check_chunk_file(base_dir, chunk_file, checks):
            verbose_print(f'broken chunk file {chunk_file}')
            return None
        d[chunk_file] = checks
    return d

def shuffle_chunk_files(base_dir:str, chunk_file:str, chunk_file2:str):
    assert chunk_file != chunk_file2
    chunks = load_chunk_file(base_dir, chunk_file)
    chunks2 = load_chunk_file(base_dir, chunk_file2)
    length = len(chunks)
    merged_chunks = chunks+chunks2
    random.shuffle(merged_chunks)
    save_chunk_file(base_dir, chunk_file, merged_chunks[:length])
    save_chunk_file(base_dir, chunk_file, merged_chunks[length:])



import regex
import random

# 正規表現パターンをコンパイル
pattern = regex.compile(r'\[(?:[^\[\]]+|(?R))*\]')

def find_nested_brackets(s):
    return regex.findall(pattern, s)

# 使用例
def da(s, mapped=None):
    mapped = mapped or {}
    matches = find_nested_brackets(s)
    for match in matches:
        if '|' not in match and '/' not in match:
            continue
        if match in mapped:
            continue
        inner = match[1:-1]
        if '[' in inner:
            inner = da(inner, mapped)
            match = f"[{inner}]"
        if '/' in inner:
            tt = inner.split('/')
            random.shuffle(tt)
            inner = ''.join(tt)
        if '|' in inner:
            tt = inner.split('|')
            random.shuffle(tt)
            inner = tt[0]
        mapped[match]=inner
    for match, inner in mapped.items():
        s = s.replace(match, inner)
    return s

s = "This is [a simple|an] example, [SEP] but [here [is [a|an] nested] [simple|small] example] for you."
print(da(s))


from .tokenizers import load_tokenizer

from .composers import (
    DataComposer,
    PretrainComposer, FinetuneComposer
)

from .t5trainers import (
    T5PretrainComposer, T5FinetuneComposer
)

"""
from papertown.papertown_tokenizer import load_tokenizer, pre_encode, post_decode
from papertown.papertown_dataset import (
    DatasetStore, DataComposer,
    FinetuneComposer, PretrainComposer,
    T5FinetuneComposer, T5PretrainComposer,
    DP
)
"""



from setuptools import setup, find_packages

"""
python3 -m unittest
vim setup.py
rm -rf dist/
python3 setup.py sdist bdist_wheel
twine upload --repository pypi dist/*
"""


def _requires_from_file(filename):
    return open(filename).read().splitlines()


setup(
    name="kogitune",
    version="0.1", #.10.14",
    license="Apache",
    author="Kimio Kuramitsu",
    description="The Kogitune 🦊 LLM Project",
    url="https://github.com/kuramitsulab/kogitune",
    packages=["kogitune"],
    package_dir={"kogitune": "kogitune"},
    package_data={"kogitune": ["*/*"]},
    install_requires=_requires_from_file("requirements.txt"),
    entry_points={
        "console_scripts": [
            "kogitune_store=kogitune.cli:main_store",
	        "kogitune_update=kogitune.cli:main_update",
        ]
    },
    classifiers=[
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Intended Audience :: Education",
    ],
)



import numpy as np
import torch 

from .tokenizers import find_extra_ids, find_newline_token_id, load_tokenizer
from .composers import DataComposer
from transformers import DataCollatorForSeq2Seq


class MSP(object):
    def __init__(self, tokenizer, lambda_=3):
        self.lambda_ = lambda_
        self.eos_token_id = tokenizer.eos_token_id
        self.extra_ids = find_extra_ids(tokenizer)
        self.newline_id = find_newline_token_id(tokenizer)

    def __call__(self, data, max_length):
        data = data.tolist()
        inputs = tokens = []
        outputs = masked = []
        start=0
        index=0
        for length in np.random.poisson(self.lambda_, 1000):
            end = start+max(1, length)
            data_part = data[start:end]
            tokens.extend(data_part)
            if self.eos_token_id in data_part or self.newline_id in data_part:
                masked.extend(data_part)
            else:
                masked.append(self.extra_ids[(index//2)%100])
                index += 1
                tokens,masked = masked, tokens
            start = end
            if start > len(data):
                break
        return {
            "input_ids": torch.tensor(inputs[:max_length//2], dtype=torch.long),
            "labels": torch.tensor(outputs[:max_length//2], dtype=torch.long),
        }


class T5PretrainComposer(DataComposer):
    def __init__(self, url_list, block_size, **kwargs):
        kwargs['training_type'] = 'pre'
        kwargs['split'] = 'train'
        DataComposer.__init__(self, url_list, block_size=block_size, **kwargs)
        tokenizer = load_tokenizer(self.tokenizer_path)
        self.build_fn = MSP(tokenizer=tokenizer)

    def get_collator(self):
        tokenizer = load_tokenizer(self.tokenizer_path)
        return DataCollatorForSeq2Seq(tokenizer)

class Seq2seq(object):

    def __call__(self, data, max_length):
        index = data[-1]
        inputs = data[:index]
        labels = data[index:-1]
        if len(inputs)+len(labels) > max_length:
            # 前半分と後ろ半分を連結する
            inputs = inputs[:(max_length -len(labels))]
        inputs[-1] = labels[-1] # eos
        return {
            "input_ids": torch.tensor(inputs.astype(np.int64), dtype=torch.long),
            "labels": torch.tensor(labels.astype(np.int64), dtype=torch.long),
        }

class T5FinetuneComposer(DataComposer):
    def __init__(self, url_list, **kwargs):
        kwargs['training_type'] = 'seq2seq'
        if 'block_size' in kwargs: # ブロックサイズは、事前学習用のパラメータ
            kwargs['max_length'] = kwargs['block_size']
            del kwargs['block_size']
        DataComposer.__init__(self, url_list, **kwargs)
        self.build_fn = Seq2seq()

    def get_collator(self):
        tokenizer = load_tokenizer(self.tokenizer_path)
        return DataCollatorForSeq2Seq(tokenizer)

    def get_eval_dataset(self):
        raise NotImplementedError()



